/*  GNU Ocrad - Optical Character Recognition program
    Copyright (C) 2003, 2004, 2005 Antonio Diaz Diaz.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <cctype>
#include <cstdio>
#include <vector>
#include "common.h"
#include "rectangle.h"
#include "vrhomboid.h"
#include "track.h"
#include "ucs.h"
#include "bitmap.h"
#include "block.h"
#include "character.h"
#include "textline.h"


Textline::Textline( const Textline & tl ) throw()
  : Track( tl ), data( tl.data )
  {
  if( tl._big_initial ) _big_initial = new Character( *(tl._big_initial) );
  else _big_initial = 0;
  }


Textline & Textline::operator=( const Textline & tl ) throw()
  {
  if( this != &tl )
    {
    Track::operator=( tl );
    delete_big_initial();
    if( tl._big_initial ) _big_initial = new Character( *(tl._big_initial) );
    data = tl.data;
    }
  return *this;
  }


void Textline::set_track() throw()
  {
  std::vector< Rectangle > rv;
  for( unsigned int i = 0; i < data.size(); ++i )
    if( !data[i].maybe(' ') ) rv.push_back( data[i] );
  Track::set_track( rv );
  }


void Textline::big_initial( const Character & c ) throw()
  {
  if( !_big_initial ) _big_initial = new Character( c );
  else *_big_initial = c;
  }


void Textline::delete_big_initial() throw()
  {
  if( _big_initial ) { delete _big_initial; _big_initial = 0; }
  }


void Textline::delete_character( int i ) throw()
  {
  if( i < 0 || i >= characters() )
    Ocrad::internal_error( "delete_character, index out of bounds" );
  data.erase( data.begin() + i );
  }


int Textline::shift_character( const Character & c ) throw()
  {
  int i = characters();

  while( i > 0 && c.h_precedes( data[i-1] ) ) --i;
  data.insert( data.begin() + i, c );
  return i;
  }


bool Textline::insert_space( int i, bool tab ) throw()
  {
  if( i <= 0 || i >= characters() )
    Ocrad::internal_error( "insert_space, index out of bounds" );
  if( height() == 0 )
    Ocrad::internal_error( "insert_space, track not set yet" );
  Character & c1 = data[i-1];
  Character & c2 = data[i];
  int l = c1.right() + 1;
  int r = c2.left() - 1;
  if( l > r ) return false;
  int t = top( ( l + r ) / 2 );
  int b = bottom( ( l + r ) / 2 );
  Character c( Rectangle( l, t, r, b ), *c1.blockmap() );
  c.add_guess( ' ', tab ? 1 : 0 );
  if( tab ) c.add_guess( '\t', 0 );
  data.insert( data.begin() + i, c );
  return true;
  }


Character & Textline::character( int i ) const throw()
  {
  if( i < 0 || i >= characters() )
    Ocrad::internal_error( "character, index out of bounds" );
  return data[i];
  }


Rectangle Textline::charbox( const Character & c ) const throw()
  {
  return Rectangle( c.left(), top( c.hcenter() ), c.right(), bottom( c.hcenter() ) );
  }


int Textline::mean_height() const throw()
  {
  int c = 0, sum = 0;

  for( int i = 0; i < characters(); ++i )
    if( !data[i].maybe(' ') ) { ++c; sum += data[i].height(); }
  if( c ) sum /= c;
  return sum;
  }


int Textline::mean_width() const throw()
  {
  int c = 0, sum = 0;

  for( int i = 0; i < characters(); ++i )
    if( !data[i].maybe(' ') ) { ++c; sum += data[i].width(); }
  if( c ) sum /= c;
  return sum;
  }


int Textline::mean_gap_width( int max_gap_width ) const throw()
  {
  int gaps = characters() - 1, sum = 0;

  for( int i = 0; i < gaps; ++i )
    {
    int gap = data[i+1].left() - data[i].right() - 1;
    if( gap > 0 && ( !max_gap_width || gap <= max_gap_width ) ) sum += gap;
    }
  if( gaps > 0 ) sum /= gaps;
  return sum;
  }


int Textline::mean_hcenter() const throw()
  {
  int sum = 0;

  for( int i = 0; i < characters(); ++i ) sum += data[i].hcenter();
  if( characters() ) sum /= characters();
  return sum;
  }


int Textline::mean_vcenter() const throw()
  {
  int sum = 0;

  for( int i = 0; i < characters(); ++i ) sum += data[i].vcenter();
  if( characters() ) sum /= characters();
  return sum;
  }


void Textline::join( Textline & tl ) throw()
  {
  if( tl._big_initial &&
      ( !_big_initial || tl._big_initial->h_precedes( *_big_initial ) ) )
    big_initial( *(tl._big_initial) );

  for( int i = 0; i < tl.characters(); ++i ) shift_character( tl.data[i] );

  tl.delete_big_initial(); tl.data.clear();
  }


void Textline::print( const Control & control ) const throw()
  {
  if( _big_initial ) _big_initial->print( control );
  for( int i = 0; i < characters(); ++i ) character( i ).print( control );
  std::fputs( "\n", control.outfile );
  }


void Textline::dprint( const Control & control, bool graph, bool recursive )
								const throw()
  {
  if( graph || recursive )
    std::fprintf( control.outfile, "mean height = %d, track segments = %d\n",
                  mean_height(), segments() );

  if( _big_initial )
    {
    Character & c = *_big_initial;
    c.dprint( control, c, graph, recursive );
    }
  for( int i = 0; i < characters(); ++i )
    {
    Character & c = character( i );
    c.dprint( control, charbox( c ), graph, recursive );
    }
  std::fputs( "\n", control.outfile );
  }


void Textline::xprint( const Control & control ) const throw()
  {
  if( _big_initial ) _big_initial->xprint( control );
  for( int i = 0; i < characters(); ++i )
    character( i ).xprint( control );
  }


void Textline::cmark( Bitmap & bitmap ) const throw()
  {
  if( _big_initial ) bitmap.draw_rectangle( *_big_initial );
  for( int i = 0; i < characters(); ++i )
    bitmap.draw_rectangle( character(i) );
  }


void Textline::recognize1( const Charset & charset ) const throw()
  {
  if( _big_initial )
    {
    Character & c = *_big_initial;
    c.recognize1( charset, c );
    if( c.guesses() == 1 )
      {
      int code = c.guess( 0 ).code;
      if( UCS::islower( code ) ) c.only_guess( UCS::toupper( code ), 0 );
      }
    else if( c.guesses() > 1 ) c.clear_guesses();
    }

  for( int i = 0; i < characters(); ++i )
    {
    Character & c = character( i );
    c.recognize1( charset, charbox( c ) );
    }
  }
